// This file is a part of Catsfoot.
// Copyright (C) 2011  Uni Research
// Copyright (C) 2011  Valentin David
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser Public License for more details.
//
// You should have received a copy of the GNU Lesser Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

#ifndef __MEMBERS_HH
# define __MEMBERS_HH

# include <utility>
# include "../type_traits/always_false.hh"

/// \defgroup Macros Macro definitions
/// \{

/// \brief Defines a wrapper for calling a member of a class
///
/// It can be useful for such a case:
/// \code
/// DEF_MEMBER_WRAPPER(foo);
/// template <typename T, ENABLE_IF(callable<member_foo(T, int)>)>
///  [...]
///    t.foo(20);
/// \endcode
# define DEF_MEMBER_WRAPPER(X...)			      \
  struct member_##X {					      \
    template <typename T, typename... U,		      \
	      typename Ret =				      \
	      decltype(std::declval<T>()		      \
		       .X(std::declval<U>()...))>	      \
    Ret operator()(T&& t, U&&... u...) const {		      \
      return std::forward<T>(t).X(std::forward<U>(u)...);     \
    }							      \
  };


/// \brief Defines a wrapper for calling a static member of a class
///
/// It can be useful for such a case:
/// \code
/// DEF_MEMBER_WRAPPER(foo);
/// template <typename T, ENABLE_IF(callable<static_member_foo<T>(int)>)>
///  [...]
///    T::foo(20);
/// \endcode
# define DEF_STATIC_MEMBER_WRAPPER(X...)		\
  template <typename T>					\
  struct static_member_##X {				\
    template <typename... U,				\
	      typename Ret =				\
	      decltype(T::X(std::declval<U>()...))>	\
    Ret operator()(U&&... u...) const {			\
      return T::X(std::forward<U>(u)...);		\
    }							\
  };

/// \brief Defines a function wrapper for calling a function.
# define DEF_FUNCTION_WRAPPER(X...)				\
  struct function_##X {						\
    template <typename... U,					\
	      typename Ret =					\
	      decltype(X(std::declval<U>()...))>		\
    Ret operator()(U&&... u...) const {				\
      return X(std::forward<U>(u)...);				\
    }								\
  };

/// \}

#endif
